// ----------------------------------------------------------------------------
// * Raima Database Manager
// *
// * Copyright (c) 2012 Raima Inc.,  All rights reserved.
// *
// * Use of this software, whether in source code format, or in executable,
// * binary object code form, is governed by the Raima LICENSE which
// * is fully described in the LICENSE.TXT file, included within this
// * distribution of files.
// * ----------------------------------------------------------------------------

#include <iostream>

// rdm.h is required in all files using the RDM Database
#include "rdm.h"

// obj13_db.h was generated from the DDL file obj13_db.ddl.
#include "obj13_db.h"

// obj13_db_api.h was generated from the DDL file obj13_db.ddl. with 
// the -cpp option
#include "obj13_db_api.h"

// Name space for the RDM C++ interface
using namespace RDM_CPP;

// EXAMPLE - obj13 for C++
// Each time you run this example:
//   1) The database is initialized (all existing contents removed)
//   2) Two Db objects are created
//   3) Each Db object opens the database is shared mode
//   4) Db1 starts a transaction, locks the office record for writes,
//      adds a few office records
//   5) Db2 attempts to read the office records (fails while task 1 has lock)
//   6) Db1 commits transaction
//   7) Db2 is able to read the office records

// The DDL (Database Definition Language) is in obj13_db.ddl. The database
// schema is embedded into the application and the database will be created
// the first time the application is run.  Subsequent runs will use the now
// existing database.

// The following RDM C++ methods are illustrated:
//      Db->Open
//      Db->Initialize
//      Db->BeginUpdate
//      Db->End
//      Db->Abort
//      Db->BeginRead
//      Db->New_recordWithFieldValues
//      Db->Get_recordsByKey
//      Cursor->AfterLast
//      Cursor->GetFieldValues
//      Cursor++
// They are fully documented in http://docs.raima.com/rdm/.

// Each of these methods can throw and an exception of type rdm_exception

// The actual database is stored in a directory named 'obj13_db' in this
// project directory.

const char * na_offices[] =
{
    "Seattle",
    "Boise",
    "San Francisco",
    "Dallas"
};

class obj13
{
public:
    Db_obj13_db    db1;
    Db_obj13_db    db2;

    void InsertData (uint16_t num_offices, const char *office_names[], Db_obj13_db& db) throw (rdm_exception&);
    void DisplayOffices (Db_obj13_db& db) throw (rdm_exception&);
    void Initialize () throw (rdm_exception&);
    void Cleanup () throw (rdm_exception&);
};

void obj13::Initialize()
{
    // Open the database and initialize it
    try
    {
        db1 = db1.Open("obj13_db");
        db1.Initialize();
    }
    catch (rdm_exception& e)
    {
        cerr << "Error " << e.error() << " (" << e.what() << ") opening/initializing obj13_db database" << endl;
        throw (e);
        return;
    }

    // Create another db object and open the database
    // Open the database and initialize it
    try
    {
        db2 = db2.Open("obj13_db");        
    }
    catch (rdm_exception& e)
    {
        cerr << "Error " << e.error() << " (" << e.what() << ") opening obj13_db database" << endl;
        throw (e);
        return;
    }
    return;
}

void obj13::InsertData(
    uint16_t      num_offices,
    const char   *office_names[],
    Db_obj13_db&  db)
{
    int16_t       ii;
    struct office office_rec;
        
    try
    {
        // Create offices in the database
        for (ii=0; ii<num_offices; ii++)
        {
            strcpy(office_rec.name, office_names[ii]);
            db.New_office_recordWithFieldValues(&office_rec);
        }
    }
    catch (rdm_exception& e)
    {
        db.Abort();
        cerr << "Error " << e.error() << " (" << e.what() << ") adding records" << endl;
        throw (e);
        return;
    }
    return;
}
    
void obj13::DisplayOffices(
    Db_obj13_db& db)
{
    struct office office_rec;
    Cursor_office office_cursor;
    recordType rtypes[]= {Cursor_office::Type (), Cursor::NoMoreTypes()};

    try
    {
        db.BeginRead(rtypes);

        // Loop through each office record and display it
        office_cursor = db.Get_office_recordsBy_name();
        while (!office_cursor.AfterLast())
        {
            office_cursor.GetFieldValues(&office_rec);
            cout << '\t' << office_rec.name << endl;
            office_cursor++;
        }
        db.End();
    }
    catch (rdm_exception& e)
    {
        cerr << "Error " << e.error() << " (" << e.what() << ") reading records" << endl;
        if(e.error() != S_UNAVAIL)
        {
            db.End();
        }
        return;
    }   
}

void obj13::Cleanup()
{
    return;
}

int32_t main(void)
{
    recordType rtypes[]= {Cursor_office::Type (), Cursor::NoMoreTypes()};
    obj13 obj;

    try
    {
        obj.Initialize();

        // Start a transaction and add NA offices
        cout << "Add offices using db1 (get write locks)" << endl;
        obj.db1.BeginUpdate(rtypes);
        obj.InsertData(RLEN(na_offices), na_offices, obj.db1);
    }
    catch (rdm_exception& e)
    {
        cerr << "Error " << e.error() << " (" << e.what() << ")" << endl;
        return EXIT_FAILURE;
    }

    try
    {
        // Display offices using db2 (should get timeout)
        cout << endl << "Displaying Offices using db2 (should timeout)." << endl;
        obj.DisplayOffices(obj.db2);
    }
    catch (rdm_exception& e)
    {
        // We expect to get this timeout value
        if(e.error() != S_UNAVAIL)
        {
            cerr << "Error " << e.error() << " (" << e.what() << ")" << endl;
            return EXIT_FAILURE;
        }
    }

    try
    {
        // Commit the transaction
        cout << endl << "Ending the transaction (will free write locks)." << endl;
        obj.db1.End();

        // Display the offices (need to get read locks)
        cout << endl << "Displaying Offices using db2 (should succeed)." << endl;
        obj.DisplayOffices(obj.db2);

        obj.Cleanup();
    }
    catch (rdm_exception& e)
    {
        cerr << "Error " << e.error() << " (" << e.what() << ")" << endl;
        return EXIT_FAILURE;
    }
    
    cout << endl << "Press <Enter> to continue . . .";
    getchar();
    return EXIT_SUCCESS;
}
